=== modified file 'NEWS'
--- NEWS	2007-04-21 15:11:39 +0000
+++ NEWS	2007-04-23 00:28:58 +0000
@@ -24,6 +24,18 @@
     * Tags are now included in logs, that use the long log formatter. 
       (Erik Bågfors, Alexander Belchenko)
 
+    * ``bzr help`` provides a clearer message when a help topic cannot be
+      found (fixes bug 107656). (Robert Collins)
+
+    * ``bzr help`` now accepts optional prefixes for command help. The help
+      for all commands can now be found at ``bzr help commands/COMMANDNAME``
+      as well as ``bzr help COMMANDNAME`` (which only works for commands 
+      where the name is not the same as a more general help topic). 
+      (Robert Collins)
+
+    * ``bzr help PLUGINNAME`` will now return the module docstring from the
+      plugin PLUGINNAME fixing bug 50408. (Robert Collins)
+
   INTERNALS:
 
     * bzrlib API compatability with 0.8 has been dropped, cleaning up some
@@ -87,6 +99,9 @@
     * Command objects can now declare related help topics by having _see_also
       set to a list of related topic. (Robert Collins)
 
+    * ``bzrlib.help`` now delegates to the Command class for Command specific
+      help. (Robert Collins)
+
   BUGFIXES:
 
     * Don't fail bundle selftest if email has 'two' embedded.  

=== modified file 'bzrlib/commands.py'
--- bzrlib/commands.py	2007-04-20 03:00:57 +0000
+++ bzrlib/commands.py	2007-04-23 00:28:58 +0000
@@ -137,6 +137,14 @@
     plugins_override
         If true, plugin commands can override builtins.
     """
+    try:
+        return _get_cmd_object(cmd_name, plugins_override)
+    except KeyError:
+        raise errors.BzrCommandError('unknown command "%s"' % cmd_name)
+
+
+def _get_cmd_object(cmd_name, plugins_override=True):
+    """Worker for get_cmd_object which raises KeyError rather than BzrCommandError."""
     from bzrlib.externalcommand import ExternalCommand
 
     # We want only 'ascii' command names, but the user may have typed
@@ -159,8 +167,7 @@
     cmd_obj = ExternalCommand.find_command(cmd_name)
     if cmd_obj:
         return cmd_obj
-
-    raise errors.BzrCommandError('unknown command "%s"' % cmd_name)
+    raise KeyError
 
 
 class Command(object):
@@ -233,14 +240,78 @@
         if self.__doc__ == Command.__doc__:
             warn("No help message set for %r" % self)
 
-    def get_see_also(self):
+    def _usage(self):
+        """Return single-line grammar for this command.
+
+        Only describes arguments, not options.
+        """
+        s = 'bzr ' + self.name() + ' '
+        for aname in self.takes_args:
+            aname = aname.upper()
+            if aname[-1] in ['$', '+']:
+                aname = aname[:-1] + '...'
+            elif aname[-1] == '?':
+                aname = '[' + aname[:-1] + ']'
+            elif aname[-1] == '*':
+                aname = '[' + aname[:-1] + '...]'
+            s += aname + ' '
+                
+        assert s[-1] == ' '
+        s = s[:-1]
+        return s
+
+    def get_help_text(self, additional_see_also=None):
+        """Return a text string with help for this command.
+        
+        :param additional_see_also: Additional help topics to be
+            cross-referenced.
+        """
+        doc = self.help()
+        if doc is None:
+            raise NotImplementedError("sorry, no detailed help yet for %r" % self.name())
+
+        result = ""
+        result += 'usage: %s\n' % self._usage()
+
+        if self.aliases:
+            result += 'aliases:\n'
+            result += ', '.join(self.aliases) + '\n'
+
+        result += '\n'
+
+        plugin_name = self.plugin_name()
+        if plugin_name is not None:
+            result += '(From plugin "%s")' % plugin_name
+            result += '\n\n'
+
+        result += doc
+        if result[-1] != '\n':
+            result += '\n'
+        result += '\n'
+        result += option.get_optparser(self.options()).format_option_help()
+        see_also = self.get_see_also(additional_see_also)
+        if see_also:
+            result += '\nSee also: '
+            result += ', '.join(see_also)
+            result += '\n'
+        return result
+
+    def get_help_topic(self):
+        """Return the commands help topic - its name."""
+        return self.name()
+
+    def get_see_also(self, additional_terms=None):
         """Return a list of help topics that are related to this ommand.
         
         The list is derived from the content of the _see_also attribute. Any
         duplicates are removed and the result is in lexical order.
+        :param additional_terms: Additional help topics to cross-reference.
         :return: A list of help topics.
         """
-        return sorted(set(getattr(self, '_see_also', [])))
+        see_also = set(getattr(self, '_see_also', []))
+        if additional_terms:
+            see_also.update(additional_terms)
+        return sorted(see_also)
 
     def options(self):
         """Return dict of valid options for this command.
@@ -288,8 +359,7 @@
             argv = []
         args, opts = parse_args(self, argv, alias_argv)
         if 'help' in opts:  # e.g. bzr add --help
-            from bzrlib.help import help_on_command
-            help_on_command(self.name())
+            sys.stdout.write(self.get_help_text())
             return 0
         # mix arguments and options into one dictionary
         cmdargs = _match_argform(self.name(), self.takes_args, args)
@@ -657,5 +727,29 @@
             pdb.post_mortem(sys.exc_traceback)
         return 3
 
+
+class HelpCommandIndex(object):
+    """A index for bzr help that returns commands."""
+
+    def __init__(self):
+        self.prefix = 'commands/'
+
+    def get_topics(self, topic):
+        """Search for topic amongst commands.
+
+        :param topic: A topic to search for.
+        :return: A list which is either empty or contains a single
+            Command entry.
+        """
+        if topic and topic.startswith(self.prefix):
+            topic = topic[len(self.prefix):]
+        try:
+            cmd = _get_cmd_object(topic)
+        except KeyError:
+            return []
+        else:
+            return [cmd]
+
+
 if __name__ == '__main__':
     sys.exit(main(sys.argv))

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2007-04-17 09:01:53 +0000
+++ bzrlib/errors.py	2007-04-23 00:28:58 +0000
@@ -223,6 +223,16 @@
     def __init__(self, revision_id):
         self.revision_id = revision_id
 
+
+class NoHelpTopic(BzrError):
+
+    _fmt = ("No help could be found for '%(topic)s'. "
+        "Please use 'bzr help topics' to obtain a list of topics.")
+
+    def __init__(self, topic):
+        self.topic = topic
+
+
 class NoSuchId(BzrError):
 
     _fmt = "The file id %(file_id)s is not present in the tree %(tree)s."
@@ -1530,6 +1540,14 @@
     _fmt = "Key %(key)s is already present in map"
 
 
+class DuplicateHelpPrefix(BzrError):
+
+    _fmt = "The prefix %(prefix)s is in the help search path twice."
+
+    def __init__(self, prefix):
+        self.prefix = prefix
+
+
 class MalformedTransform(BzrError):
 
     _fmt = "Tree transform is malformed %(conflicts)r"

=== modified file 'bzrlib/help.py'
--- bzrlib/help.py	2007-04-19 04:54:08 +0000
+++ bzrlib/help.py	2007-04-23 00:24:53 +0000
@@ -25,8 +25,12 @@
 import textwrap
 
 from bzrlib import (
+    commands as _mod_commands,
+    errors,
     help_topics,
     osutils,
+    plugin,
+    symbol_versioning,
     )
 
 
@@ -35,101 +39,12 @@
     if outfile is None:
         outfile = sys.stdout
 
-    if topic is None:
-        topic = 'basic'
-
-    if topic in help_topics.topic_registry:
-        txt = help_topics.topic_registry.get_detail(topic)
-        outfile.write(txt)
-    else:
-        help_on_command(topic, outfile=outfile)
-
-
-def command_usage(cmd_object):
-    """Return single-line grammar for command.
-
-    Only describes arguments, not options.
-    """
-    s = 'bzr ' + cmd_object.name() + ' '
-    for aname in cmd_object.takes_args:
-        aname = aname.upper()
-        if aname[-1] in ['$', '+']:
-            aname = aname[:-1] + '...'
-        elif aname[-1] == '?':
-            aname = '[' + aname[:-1] + ']'
-        elif aname[-1] == '*':
-            aname = '[' + aname[:-1] + '...]'
-        s += aname + ' '
-            
-    assert s[-1] == ' '
-    s = s[:-1]
-    
-    return s
-
-
-def print_command_plugin(cmd_object, outfile, format):
-    """Print the plugin that provides a command object, if any.
-
-    If the cmd_object is provided by a plugin, prints the plugin name to
-    outfile using the provided format string.
-    """
-    plugin_name = cmd_object.plugin_name()
-    if plugin_name is not None:
-        out_str = '(From plugin "%s")' % plugin_name
-        outfile.write(format % out_str)
-
-
-def help_on_command(cmdname, outfile=None):
-    from bzrlib.commands import get_cmd_object
-
-    cmdname = str(cmdname)
-    cmd_object = get_cmd_object(cmdname)
-
-    return help_on_command_object(cmd_object, cmdname, outfile)
-
-
-def help_on_command_object(cmd_object, cmdname, outfile=None):
-    """Generate help on the cmd_object with a supplied name of cmdname.
-
-    :param cmd_object: An instance of a Command.
-    :param cmdname: The user supplied name. This might be an alias for example.
-    :param outfile: A stream to write the help to.
-    """
-    if outfile is None:
-        outfile = sys.stdout
-
-    doc = cmd_object.help()
-    if doc is None:
-        raise NotImplementedError("sorry, no detailed help yet for %r" % cmdname)
-
-    print >>outfile, 'usage:', command_usage(cmd_object)
-
-    if cmd_object.aliases:
-        print >>outfile, 'aliases:',
-        print >>outfile, ', '.join(cmd_object.aliases)
-
-    print >>outfile
-
-    print_command_plugin(cmd_object, outfile, '%s\n\n')
-
-    outfile.write(doc)
-    if doc[-1] != '\n':
-        outfile.write('\n')
-    help_on_command_options(cmd_object, outfile)
-    see_also = cmd_object.get_see_also()
-    if see_also:
-        outfile.write('\nSee also: ')
-        outfile.write(', '.join(see_also))
-        outfile.write('\n')
-
-
-def help_on_command_options(cmd, outfile=None):
-    from bzrlib.option import Option, get_optparser
-    if outfile is None:
-        outfile = sys.stdout
-    options = cmd.options()
-    outfile.write('\n')
-    outfile.write(get_optparser(options).format_option_help())
+    indices = HelpIndices()
+    topics = indices.search(topic)
+    shadowed_terms = []
+    for index, topic in topics[1:]:
+        shadowed_terms.append('%s%s' % (index.prefix, topic.get_help_topic()))
+    outfile.write(topics[0][1].get_help_text(shadowed_terms))
 
 
 def help_commands(outfile=None):
@@ -139,19 +54,21 @@
     outfile.write(_help_commands_to_text('commands'))
 
 
+@symbol_versioning.deprecated_function(symbol_versioning.zero_sixteen)
+def command_usage(cmd):
+    return cmd._usage()
+
+
 def _help_commands_to_text(topic):
     """Generate the help text for the list of commands"""
-    from bzrlib.commands import (builtin_command_names,
-                                 plugin_command_names,
-                                 get_cmd_object)
     out = []
     if topic == 'hidden-commands':
         hidden = True
     else:
         hidden = False
-    names = set(builtin_command_names()) # to eliminate duplicates
-    names.update(plugin_command_names())
-    commands = ((n, get_cmd_object(n)) for n in names)
+    names = set(_mod_commands.builtin_command_names()) # to eliminate duplicates
+    names.update(_mod_commands.plugin_command_names())
+    commands = ((n, _mod_commands.get_cmd_object(n)) for n in names)
     shown_commands = [(n, o) for n, o in commands if o.hidden == hidden]
     max_name = max(len(n) for n, o in shown_commands)
     indent = ' ' * (max_name + 1)
@@ -180,6 +97,54 @@
 help_topics.topic_registry.register("commands",
                                     _help_commands_to_text,
                                     "Basic help for all commands")
-help_topics.topic_registry.register("hidden-commands", 
+help_topics.topic_registry.register("hidden-commands",
                                     _help_commands_to_text,
                                     "All hidden commands")
+
+
+class HelpIndices(object):
+    """Maintainer of help topics across multiple indices.
+    
+    It is currently separate to the HelpTopicRegistry because of its ordered
+    nature, but possibly we should instead structure it as a search within the
+    registry and add ordering and searching facilities to the registry. The
+    registry would probably need to be restructured to support that cleanly
+    which is why this has been implemented in parallel even though it does as a
+    result permit searching for help in indices which are not discoverable via
+    'help topics'.
+
+    Each index has a unique prefix string, such as "commands", and contains
+    help topics which can be listed or searched.
+    """
+
+    def __init__(self):
+        self.search_path = [
+            help_topics.HelpTopicIndex(),
+            _mod_commands.HelpCommandIndex(),
+            plugin.PluginsHelpIndex(),
+            ]
+
+    def _check_prefix_uniqueness(self):
+        """Ensure that the index collection is able to differentiate safely."""
+        prefixes = {}
+        for index in self.search_path:
+            prefixes.setdefault(index.prefix, []).append(index)
+        for prefix, indices in prefixes.items():
+            if len(indices) > 1:
+                raise errors.DuplicateHelpPrefix(prefix)
+
+    def search(self, topic):
+        """Search for topic across the help search path.
+        
+        :param topic: A string naming the help topic to search for.
+        :raises: NoHelpTopic if none of the indexs in search_path have topic.
+        :return: A list of HelpTopics which matched 'topic'.
+        """
+        self._check_prefix_uniqueness()
+        result = []
+        for index in self.search_path:
+            result.extend([(index, _topic) for _topic in index.get_topics(topic)])
+        if not result:
+            raise errors.NoHelpTopic(topic)
+        else:
+            return result

=== modified file 'bzrlib/help_topics.py'
--- bzrlib/help_topics.py	2007-04-01 15:52:43 +0000
+++ bzrlib/help_topics.py	2007-04-20 05:36:58 +0000
@@ -243,3 +243,62 @@
 topic_registry.register('checkouts', _checkouts,
                         'Information on what a checkout is')
 
+
+class HelpTopicIndex(object):
+    """A index for bzr help that returns topics."""
+
+    def __init__(self):
+        self.prefix = ''
+
+    def get_topics(self, topic):
+        """Search for topic in the HelpTopicRegistry.
+
+        :param topic: A topic to search for. None is treated as 'basic'.
+        :return: A list which is either empty or contains a single
+            RegisteredTopic entry.
+        """
+        if topic is None:
+            topic = 'basic'
+        if topic in topic_registry:
+            return [RegisteredTopic(topic)]
+        else:
+            return []
+
+
+class RegisteredTopic(object):
+    """A help topic which has been registered in the HelpTopicRegistry.
+
+    These topics consist of nothing more than the name of the topic - all
+    data is retrieved on demand from the registry.
+    """
+
+    def __init__(self, topic):
+        """Constructor.
+
+        :param topic: The name of the topic that this represents.
+        """
+        self.topic = topic
+
+    def get_help_text(self, additional_see_also=None):
+        """Return a string with the help for this topic.
+
+        :param additional_see_also: Additional help topics to be
+            cross-referenced.
+        """
+        result = topic_registry.get_detail(self.topic)
+        # there is code duplicated here and in bzrlib/plugin.py's 
+        # matching Topic code. This should probably be factored in
+        # to a helper function and a common base class.
+        if additional_see_also is not None:
+            see_also = sorted(set(additional_see_also))
+        else:
+            see_also = None
+        if see_also:
+            result += '\nSee also: '
+            result += ', '.join(see_also)
+            result += '\n'
+        return result
+
+    def get_help_topic(self):
+        """Return the help topic this can be found under."""
+        return self.topic

=== modified file 'bzrlib/plugin.py'
--- bzrlib/plugin.py	2007-02-23 06:06:54 +0000
+++ bzrlib/plugin.py	2007-04-20 05:44:37 +0000
@@ -267,3 +267,71 @@
             warning('Unable to load plugin %r from %r'
                     % (name, zip_name))
             log_exception_quietly()
+
+
+class PluginsHelpIndex(object):
+    """A help index that returns help topics for plugins."""
+
+    def __init__(self):
+        self.prefix = 'plugins/'
+
+    def get_topics(self, topic):
+        """Search for topic in the loaded plugins.
+
+        This will not trigger loading of new plugins.
+
+        :param topic: A topic to search for.
+        :return: A list which is either empty or contains a single
+            RegisteredTopic entry.
+        """
+        if not topic:
+            return []
+        if topic.startswith(self.prefix):
+            topic = topic[len(self.prefix):]
+        plugin_module_name = 'bzrlib.plugins.%s' % topic
+        try:
+            module = sys.modules[plugin_module_name]
+        except KeyError:
+            return []
+        else:
+            return [ModuleHelpTopic(module)]
+
+
+class ModuleHelpTopic(object):
+    """A help topic which returns the docstring for a module."""
+
+    def __init__(self, module):
+        """Constructor.
+
+        :param module: The module for which help should be generated.
+        """
+        self.module = module
+
+    def get_help_text(self, additional_see_also=None):
+        """Return a string with the help for this topic.
+
+        :param additional_see_also: Additional help topics to be
+            cross-referenced.
+        """
+        if not self.module.__doc__:
+            result = "Plugin '%s' has no docstring.\n" % self.module.__name__
+        else:
+            result = self.module.__doc__
+        if result[-1] != '\n':
+            result += '\n'
+        # there is code duplicated here and in bzrlib/help_topic.py's 
+        # matching Topic code. This should probably be factored in
+        # to a helper function and a common base class.
+        if additional_see_also is not None:
+            see_also = sorted(set(additional_see_also))
+        else:
+            see_also = None
+        if see_also:
+            result += 'See also: '
+            result += ', '.join(see_also)
+            result += '\n'
+        return result
+
+    def get_help_topic(self):
+        """Return the modules help topic - its __name__ after bzrlib.plugins.."""
+        return self.module.__name__[len('bzrlib.plugins.'):]

=== modified file 'bzrlib/tests/blackbox/test_too_much.py'
--- bzrlib/tests/blackbox/test_too_much.py	2006-11-05 14:22:38 +0000
+++ bzrlib/tests/blackbox/test_too_much.py	2007-04-23 01:41:05 +0000
@@ -434,7 +434,7 @@
         runbzr("help slartibartfast", 3)
 
         out = capture("help ci")
-        out.index('aliases: ')
+        out.index('aliases:\n')
 
         f = file('hello.txt', 'wt')
         f.write('some nice new content\n')

=== modified file 'bzrlib/tests/test_commands.py'
--- bzrlib/tests/test_commands.py	2007-04-18 08:39:02 +0000
+++ bzrlib/tests/test_commands.py	2007-04-20 03:54:06 +0000
@@ -125,3 +125,12 @@
             _see_also = ['foo', 'bar']
         command = ACommand()
         self.assertEqual(['bar', 'foo'], command.get_see_also())
+
+    def test_additional_terms(self):
+        """Additional terms can be supplied and are deduped and sorted."""
+        class ACommand(commands.Command):
+            _see_also = ['foo', 'bar']
+        command = ACommand()
+        self.assertEqual(['bar', 'foo', 'gam'],
+            command.get_see_also(['gam', 'bar', 'gam']))
+

=== modified file 'bzrlib/tests/test_errors.py'
--- bzrlib/tests/test_errors.py	2007-04-17 09:01:53 +0000
+++ bzrlib/tests/test_errors.py	2007-04-23 00:28:58 +0000
@@ -40,6 +40,11 @@
         self.assertEqualDiff('File id {a_file_id} already exists in inventory'
                              ' as foo', str(error))
 
+    def test_duplicate_help_prefix(self):
+        error = errors.DuplicateHelpPrefix('foo')
+        self.assertEqualDiff('The prefix foo is in the help search path twice.',
+            str(error))
+
     def test_inventory_modified(self):
         error = errors.InventoryModified("a tree to be repred")
         self.assertEqualDiff("The current inventory for the tree 'a tree to "
@@ -91,6 +96,12 @@
             "smart protocol.",
             str(error))
 
+    def test_no_help_topic(self):
+        error = errors.NoHelpTopic("topic")
+        self.assertEqualDiff("No help could be found for 'topic'. "
+            "Please use 'bzr help topics' to obtain a list of topics.",
+            str(error))
+
     def test_no_such_id(self):
         error = errors.NoSuchId("atree", "anid")
         self.assertEqualDiff("The file id anid is not present in the tree "

=== modified file 'bzrlib/tests/test_help.py'
--- bzrlib/tests/test_help.py	2007-04-19 07:40:42 +0000
+++ bzrlib/tests/test_help.py	2007-04-23 00:24:53 +0000
@@ -19,8 +19,12 @@
 from cStringIO import StringIO
 
 from bzrlib import (
+    builtins,
     commands,
+    errors,
     help,
+    help_topics,
+    plugin,
     tests,
     )
 
@@ -33,10 +37,222 @@
             """A sample command."""
             _see_also = ['foo', 'bar']
         cmd = cmd_WithSeeAlso()
-        helpfile = StringIO()
-        help.help_on_command_object(cmd, 'cmd_sample', helpfile)
-        self.assertEndsWith(
-            helpfile.getvalue(),
-            '  -h, --help  show help message\n'
-            '\n'
-            'See also: bar, foo\n')
+        helptext = cmd.get_help_text()
+        self.assertEndsWith(
+            helptext,
+            '  -h, --help  show help message\n'
+            '\n'
+            'See also: bar, foo\n')
+
+    def test_get_help_text(self):
+        """Commands have a get_help_text method which returns their help."""
+        class cmd_Demo(commands.Command):
+            """A sample command."""
+        cmd = cmd_Demo()
+        helptext = cmd.get_help_text()
+        self.assertStartsWith(helptext, 'usage: bzr Demo')
+        self.assertEndsWith(helptext, 'show help message\n')
+
+    def test_command_with_additional_see_also(self):
+        class cmd_WithSeeAlso(commands.Command):
+            """A sample command."""
+            _see_also = ['foo', 'bar']
+        cmd = cmd_WithSeeAlso()
+        helptext = cmd.get_help_text(['gam'])
+        self.assertEndsWith(
+            helptext,
+            '  -h, --help  show help message\n'
+            '\n'
+            'See also: bar, foo, gam\n')
+
+    def test_command_only_additional_see_also(self):
+        class cmd_WithSeeAlso(commands.Command):
+            """A sample command."""
+        cmd = cmd_WithSeeAlso()
+        helptext = cmd.get_help_text(['gam'])
+        self.assertEndsWith(
+            helptext,
+            '  -h, --help  show help message\n'
+            '\n'
+            'See also: gam\n')
+
+    def test_get_help_topic(self):
+        """The help topic for a Command is its name()."""
+        class cmd_foo_bar(commands.Command):
+            """A sample command."""
+        cmd = cmd_foo_bar()
+        self.assertEqual(cmd.name(), cmd.get_help_topic())
+    
+
+class TestRegisteredTopic(tests.TestCase):
+    """Tests for the RegisteredTopic class."""
+
+    def test_contruct(self):
+        """Construction takes the help topic name for the registered item."""
+        # validate our test 
+        self.assertTrue('basic' in help_topics.topic_registry)
+        topic = help_topics.RegisteredTopic('basic')
+        self.assertEqual('basic', topic.topic)
+
+    def test_get_help_text(self):
+        """A RegisteredTopic returns the get_detail results for get_help_text."""
+        topic = help_topics.RegisteredTopic('commands')
+        self.assertEqual(help_topics.topic_registry.get_detail('commands'),
+            topic.get_help_text())
+
+    def test_get_help_text_with_additional_see_also(self):
+        topic = help_topics.RegisteredTopic('commands')
+        self.assertEndsWith(
+            topic.get_help_text(['foo', 'bar']),
+            '\n'
+            'See also: bar, foo\n')
+
+    def test_get_help_topic(self):
+        """The help topic for a RegisteredTopic is its topic from construction."""
+        topic = help_topics.RegisteredTopic('foobar')
+        self.assertEqual('foobar', topic.get_help_topic())
+        topic = help_topics.RegisteredTopic('baz')
+        self.assertEqual('baz', topic.get_help_topic())
+
+
+class TestTopicIndex(tests.TestCase):
+    """Tests for the HelpTopicIndex class."""
+
+    def test_default_constructable(self):
+        index = help_topics.HelpTopicIndex()
+
+    def test_get_topics_None(self):
+        """Searching for None returns the basic help topic."""
+        index = help_topics.HelpTopicIndex()
+        topics = index.get_topics(None)
+        self.assertEqual(1, len(topics))
+        self.assertIsInstance(topics[0], help_topics.RegisteredTopic)
+        self.assertEqual('basic', topics[0].topic)
+
+    def test_get_topics_topics(self):
+        """Searching for a string returns the matching string."""
+        index = help_topics.HelpTopicIndex()
+        topics = index.get_topics('topics')
+        self.assertEqual(1, len(topics))
+        self.assertIsInstance(topics[0], help_topics.RegisteredTopic)
+        self.assertEqual('topics', topics[0].topic)
+
+    def test_get_topics_no_topic(self):
+        """Searching for something not registered returns []."""
+        index = help_topics.HelpTopicIndex()
+        self.assertEqual([], index.get_topics('nothing by this name'))
+
+    def test_prefix(self):
+        """TopicIndex has a prefix of ''."""
+        index = help_topics.HelpTopicIndex()
+        self.assertEqual('', index.prefix)
+
+
+class TestCommandIndex(tests.TestCase):
+    """Tests for the HelpCommandIndex class."""
+
+    def test_default_constructable(self):
+        index = commands.HelpCommandIndex()
+
+    def test_get_topics_None(self):
+        """Searching for None returns an empty list."""
+        index = commands.HelpCommandIndex()
+        self.assertEqual([], index.get_topics(None))
+
+    def test_get_topics_rocks(self):
+        """Searching for 'rocks' returns the cmd_rocks command instance."""
+        index = commands.HelpCommandIndex()
+        topics = index.get_topics('rocks')
+        self.assertEqual(1, len(topics))
+        self.assertIsInstance(topics[0], builtins.cmd_rocks)
+
+    def test_get_topics_no_topic(self):
+        """Searching for something that is not a command returns []."""
+        index = commands.HelpCommandIndex()
+        self.assertEqual([], index.get_topics('nothing by this name'))
+
+    def test_prefix(self):
+        """CommandIndex has a prefix of 'commands/'."""
+        index = commands.HelpCommandIndex()
+        self.assertEqual('commands/', index.prefix)
+
+    def test_get_topic_with_prefix(self):
+        """Searching for commands/rocks returns the rocks command object."""
+        index = commands.HelpCommandIndex()
+        topics = index.get_topics('commands/rocks')
+        self.assertEqual(1, len(topics))
+        self.assertIsInstance(topics[0], builtins.cmd_rocks)
+
+
+class TestHelpIndices(tests.TestCase):
+    """Tests for the HelpIndices class."""
+
+    def test_default_search_path(self):
+        """The default search path should include internal indexs."""
+        indices = help.HelpIndices()
+        self.assertEqual(3, len(indices.search_path))
+        # help topics should be searched in first.
+        self.assertIsInstance(indices.search_path[0],
+            help_topics.HelpTopicIndex)
+        # with commands being search second.
+        self.assertIsInstance(indices.search_path[1],
+            commands.HelpCommandIndex)
+        # and plugins are a third index.
+        self.assertIsInstance(indices.search_path[2],
+            plugin.PluginsHelpIndex)
+
+    def test_search_for_unknown_topic_raises(self):
+        """Searching for an unknown topic should raise NoHelpTopic."""
+        indices = help.HelpIndices()
+        indices.search_path = []
+        error = self.assertRaises(errors.NoHelpTopic, indices.search, 'foo')
+        self.assertEqual('foo', error.topic)
+
+    def test_search_calls_get_topic(self):
+        """Searching should call get_topics in all indexes in order."""
+        calls = []
+        class RecordingIndex(object):
+            def __init__(self, name):
+                self.prefix = name
+            def get_topics(self, topic):
+                calls.append(('get_topics', self.prefix, topic))
+                return ['something']
+        index = help.HelpIndices()
+        index.search_path = [RecordingIndex('1'), RecordingIndex('2')]
+        # try with None
+        index.search(None)
+        self.assertEqual([
+            ('get_topics', '1', None),
+            ('get_topics', '2', None),
+            ],
+            calls)
+        # and with a string
+        del calls[:]
+        index.search('bar')
+        self.assertEqual([
+            ('get_topics', '1', 'bar'),
+            ('get_topics', '2', 'bar'),
+            ],
+            calls)
+
+    def test_search_returns_index_and_results(self):
+        """Searching should return help topics with their index"""
+        class CannedIndex(object):
+            def __init__(self, prefix, search_result):
+                self.prefix = prefix
+                self.result = search_result
+            def get_topics(self, topic):
+                return self.result
+        index = help.HelpIndices()
+        index_one = CannedIndex('1', ['a'])
+        index_two = CannedIndex('2', ['b', 'c'])
+        index.search_path = [index_one, index_two]
+        self.assertEqual([(index_one, 'a'), (index_two, 'b'), (index_two, 'c')],
+            index.search(None))
+
+    def test_search_checks_for_duplicate_prefixes(self):
+        """Its an error when there are multiple indices with the same prefix."""
+        indices = help.HelpIndices()
+        indices.search_path = [help_topics.HelpTopicIndex(),
+            help_topics.HelpTopicIndex()]
+        self.assertRaises(errors.DuplicateHelpPrefix, indices.search, None)

=== modified file 'bzrlib/tests/test_plugins.py'
--- bzrlib/tests/test_plugins.py	2007-02-02 16:17:55 +0000
+++ bzrlib/tests/test_plugins.py	2007-04-20 05:44:37 +0000
@@ -25,6 +25,7 @@
 import sys
 import zipfile
 
+from bzrlib import plugin, tests
 import bzrlib.plugin
 import bzrlib.plugins
 import bzrlib.commands
@@ -160,17 +161,15 @@
         for cmd_name in bzrlib.commands.builtin_command_names():
             if cmd_name in bzrlib.commands.plugin_command_names():
                 continue
-            help = StringIO()
             try:
-                bzrlib.help.help_on_command(cmd_name, help)
+                help = bzrlib.commands.get_cmd_object(cmd_name).get_help_text()
             except NotImplementedError:
                 # some commands have no help
                 pass
             else:
-                help.seek(0)
-                self.assertNotContainsRe(help.read(), 'From plugin "[^"]*"')
+                self.assertNotContainsRe(help, 'From plugin "[^"]*"')
 
-            if help in help_commands.keys():
+            if cmd_name in help_commands.keys():
                 # some commands are hidden
                 help = help_commands[cmd_name]
                 self.assertNotContainsRe(help, 'From plugin "[^"]*"')
@@ -238,3 +237,105 @@
             self.assertEqual(expected_path, bzrlib.plugins.__path__)
         finally:
             bzrlib.plugins.__path__ = old_path
+
+
+class TestHelpIndex(tests.TestCase):
+    """Tests for the PluginsHelpIndex class."""
+
+    def test_default_constructable(self):
+        index = plugin.PluginsHelpIndex()
+
+    def test_get_topics_None(self):
+        """Searching for None returns an empty list."""
+        index = plugin.PluginsHelpIndex()
+        self.assertEqual([], index.get_topics(None))
+
+    def test_get_topics_launchpad(self):
+        """Searching for 'launchpad' returns the launchpad plugin docstring."""
+        index = plugin.PluginsHelpIndex()
+        # if bzr was run with '--no-plugins' we need to manually load the
+        # reference plugin. Its shipped with bzr, and loading at this point
+        # won't add additional tests to run.
+        import bzrlib.plugins.launchpad
+        topics = index.get_topics('launchpad')
+        self.assertEqual(1, len(topics))
+        self.assertIsInstance(topics[0], plugin.ModuleHelpTopic)
+        self.assertEqual(bzrlib.plugins.launchpad, topics[0].module)
+
+    def test_get_topics_no_topic(self):
+        """Searching for something that is not a plugin returns []."""
+        # test this by using a name that cannot be a plugin - its not
+        # a valid python identifier.
+        index = plugin.PluginsHelpIndex()
+        self.assertEqual([], index.get_topics('nothing by this name'))
+
+    def test_prefix(self):
+        """PluginsHelpIndex has a prefix of 'plugins/'."""
+        index = plugin.PluginsHelpIndex()
+        self.assertEqual('plugins/', index.prefix)
+
+    def test_get_topic_with_prefix(self):
+        """Searching for plugins/launchpad returns launchpad module help."""
+        index = plugin.PluginsHelpIndex()
+        # if bzr was run with '--no-plugins' we need to manually load the
+        # reference plugin. Its shipped with bzr, and loading at this point
+        # won't add additional tests to run.
+        import bzrlib.plugins.launchpad
+        topics = index.get_topics('plugins/launchpad')
+        self.assertEqual(1, len(topics))
+        self.assertIsInstance(topics[0], plugin.ModuleHelpTopic)
+        self.assertEqual(bzrlib.plugins.launchpad, topics[0].module)
+
+
+class FakeModule(object):
+    """A fake module to test with."""
+
+    def __init__(self, doc, name):
+        self.__doc__ = doc
+        self.__name__ = name
+
+
+class TestModuleHelpTopic(tests.TestCase):
+    """Tests for the ModuleHelpTopic class."""
+
+    def test_contruct(self):
+        """Construction takes the module to document."""
+        mod = FakeModule('foo', 'foo')
+        topic = plugin.ModuleHelpTopic(mod)
+        self.assertEqual(mod, topic.module)
+
+    def test_get_help_text_None(self):
+        """A ModuleHelpTopic returns the docstring for get_help_text."""
+        mod = FakeModule(None, 'demo')
+        topic = plugin.ModuleHelpTopic(mod)
+        self.assertEqual("Plugin 'demo' has no docstring.\n",
+            topic.get_help_text())
+
+    def test_get_help_text_no_carriage_return(self):
+        """ModuleHelpTopic.get_help_text adds a \n if needed."""
+        mod = FakeModule('one line of help', 'demo')
+        topic = plugin.ModuleHelpTopic(mod)
+        self.assertEqual("one line of help\n",
+            topic.get_help_text())
+
+    def test_get_help_text_carriage_return(self):
+        """ModuleHelpTopic.get_help_text adds a \n if needed."""
+        mod = FakeModule('two lines of help\nand more\n', 'demo')
+        topic = plugin.ModuleHelpTopic(mod)
+        self.assertEqual("two lines of help\nand more\n",
+            topic.get_help_text())
+
+    def test_get_help_text_with_additional_see_also(self):
+        mod = FakeModule('two lines of help\nand more', 'demo')
+        topic = plugin.ModuleHelpTopic(mod)
+        self.assertEqual("two lines of help\nand more\nSee also: bar, foo\n",
+            topic.get_help_text(['foo', 'bar']))
+
+    def test_get_help_topic(self):
+        """The help topic for a plugin is its module name."""
+        mod = FakeModule('two lines of help\nand more', 'bzrlib.plugins.demo')
+        topic = plugin.ModuleHelpTopic(mod)
+        self.assertEqual('demo', topic.get_help_topic())
+        mod = FakeModule('two lines of help\nand more', 'bzrlib.plugins.foo_bar')
+        topic = plugin.ModuleHelpTopic(mod)
+        self.assertEqual('foo_bar', topic.get_help_topic())

=== modified file 'tools/doc_generate/autodoc_man.py'
--- tools/doc_generate/autodoc_man.py	2007-04-19 05:32:24 +0000
+++ tools/doc_generate/autodoc_man.py	2007-04-23 01:41:05 +0000
@@ -80,7 +80,7 @@
         cmd_help = cmd_object.help()
         if cmd_help:
             firstline = cmd_help.split('\n', 1)[0]
-            usage = bzrlib.help.command_usage(cmd_object)
+            usage = cmd_object._usage()
             tmp = '.TP\n.B "%s"\n%s\n' % (usage, firstline)
             output = output + tmp
         else:
@@ -101,7 +101,7 @@
 
 def format_command (params, cmd):
     """Provides long help for each public command"""
-    subsection_header = '.SS "%s"\n' % (bzrlib.help.command_usage(cmd))
+    subsection_header = '.SS "%s"\n' % (cmd._usage())
     doc = "%s\n" % (cmd.__doc__)
     doc = cmd.help()
 

=== modified file 'tools/doc_generate/autodoc_rstx.py'
--- tools/doc_generate/autodoc_rstx.py	2007-04-19 05:32:24 +0000
+++ tools/doc_generate/autodoc_rstx.py	2007-04-23 01:41:05 +0000
@@ -73,7 +73,7 @@
         cmd_help = cmd_object.help()
         if cmd_help:
             firstline = cmd_help.split('\n', 1)[0]
-            usage = bzrlib.help.command_usage(cmd_object)
+            usage = cmd_object._usage()
             tmp = '**%s**\n\t%s\n\n' % (usage, firstline)
             output = output + tmp
         else:
@@ -97,7 +97,7 @@
 
 def format_command (params, cmd, name):
     """Provides long help for each public command"""
-    usage = bzrlib.help.command_usage(cmd)
+    usage = cmd._usage()
     subsection_header = """
 %s
 %s

